/////////////////////////////////////////////////////////////////////
//
//	Applied Data Systems, Inc
//
//  ADSparse.cpp : includes function to parse the Registry file
//  710020-1121x
//
//  Description
//  -----------
//  ADSReg.exe is a registry file parser.  It takes a single
//
//  Feb 26, 03 : ctacke
//   - added support for REG_BINARY values
//  Feb 27, 03 : ctacke
//   - added support for REG_EXPAND_SZ and REG_MULTI_SZ values
//  (710020-11214) Mar 4, 03 : ctacke
//   - added support for quotes within a REG_SZ value
//  (710020-11215) Mar 20, 03 : ctacke 
//   - added RTrim and TrimComment functions
//   - added preprocessing when getting lines to trim comments and white spaces
//   - tested and fixed comment/quote parsing issues
//   - added support for "multi_sz" option for REG_MULTI_SZ values
//
//
//  Parsing Rules
//  -------------
//  String (REG_SZ) values: Lines with a string value are delimited by the first
//            and last double quote on the line.  This allows any number of 
//            double quotes to be cotained within a string.  To include a double
//            quote do _not_ use an escape sequence (\).
//  Comments: Commensts are designated by a semicolon (;).  Anything after
//            a semicolon in a line is ignored _except_ for semicolons inside
//            a valid string value (see above).  Also, double quotes are not
//            allowed inside a comment as the String rule takes precedent.
//  Line continuations:  Some lines, especially REG_BINARY (hex) values will often
//            be many lines long.  To signify that a value continues on the next line
//            terminate the line with a backslash (\) character.  Comments are
//            allowed after the continuation character and will be ignored.
// REG_MULTI_SZ: Multiple line strings are handled in 2 possible variations. First
//            standard registry editors will export the key as a hex(7) value 
//            followed by comma delimited ascii values.  Each separate line must
//            be separated by 1 and only 1 null (ascii 00).  Alternately you can use
//            "multi_sz" as a type descriptor followed by quoted string values.
//            Each line must be separated by a comma outside the values' delimiting
//            quotes.  There must be no spaces between the comma and the preceeding
//            quote.
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include "ADSFile.h"
#include "ADSparse.h"

#ifndef ADS_DEBUG
#define ADS_DEBUG FALSE
#endif

#define MAX_CAHRS_LINE	1024

void RTrim(TCHAR *szVal);
void TrimComment(TCHAR *szVal);

///////////////////////////////////////////////////////////////////////
//////////////////////  class CValueEntry  ////////////////////////////
///////////////////////////////////////////////////////////////////////
CValueEntry::CValueEntry(TCHAR *pValName, DWORD type, void *pData, DWORD	DataSize)
{
	m_pValName = NULL;
	m_DataSize = 0;
	m_pNext = NULL;
	if (pValName)
	{
		m_pValName = (TCHAR *)new TCHAR[_tcslen(pValName)+1];
		_tcscpy(m_pValName, pValName);

		m_Type = type;

		if (DataSize && pData)
		{
			m_pData = new BYTE[DataSize];
			memcpy(m_pData, pData, (size_t)DataSize);
			m_DataSize = DataSize;
		}
	}
}

CValueEntry::~CValueEntry()
{
	if (m_pValName)
		delete [] m_pValName;
	if (m_pData)
		delete [] m_pData;
}


///////////////////////////////////////////////////////////////////////
//////////////////////  class CKeyEntry  //////////////////////////////
///////////////////////////////////////////////////////////////////////
CKeyEntry::CKeyEntry(TCHAR *pKeyName)
{
	if (!pKeyName) 
		m_pKeyName = pKeyName;
	else
	{
		m_pKeyName = (TCHAR *)new TCHAR[_tcslen(pKeyName)+1];
		_tcscpy(m_pKeyName, pKeyName);

		m_pValueList = NULL;
	}

	m_pNext = NULL;
}

CKeyEntry::~CKeyEntry()
{
	if (m_pKeyName)
		delete [] m_pKeyName;

	// delete Value list
	CValueEntry *pve;

	while(m_pValueList)
	{
		pve = m_pValueList->m_pNext;
		delete m_pValueList;
		m_pValueList = pve;
	}
}


void CKeyEntry::InsertValue(CValueEntry *pValue)
{
	CValueEntry *pPrev, *pCurr;

	// find end of the list

	pCurr = pPrev = m_pValueList;
	while(pCurr)
	{
		pPrev = pCurr;
		pCurr = pCurr->m_pNext;
	}

	if (pPrev == NULL)
	{// the first entry
		m_pValueList = pValue;
	}
	else
		pPrev->m_pNext = pValue;
}


///////////////////////////////////////////////////////////////////////
//////////////////////  class CADSparse  //////////////////////////////
///////////////////////////////////////////////////////////////////////

CADSparse::CADSparse(ADSFile *pFile)
{
	 m_pKeyEntryList = NULL;
	 m_pFile = pFile;
}

CADSparse::~CADSparse()
{CKeyEntry *pnext;

	while (m_pKeyEntryList)
	{
		pnext = m_pKeyEntryList->m_pNext;
		delete m_pKeyEntryList;
		m_pKeyEntryList = pnext;
	}

}

// pstr - Null-terminated searched string
// pLex - output lex if return is TRUE
// EndInd - index for new search relative pstr
BOOL CADSparse::GetLex(TCHAR *pstr, TCHAR *pLex, int& EndInd)
{
	TCHAR *pToken;
 	TCHAR Set1[] = _T(";[\"=:");

	TCHAR SetK[] = _T("]");
 	TCHAR SetN[] = _T("\"");
 	TCHAR EndL;
 	size_t ret;
 	TCHAR *pbeg;

	ret = _tcscspn( pstr, Set1);
	
	if (ret >= _tcslen(pstr)) 
		return FALSE;

	EndInd = 0;

	switch (pstr[ret])
	{
	case ';': 
			return FALSE;
		break;
	case '[':
			pbeg = pstr + ret;
			EndL = ']';
			EndInd = ret;

			// find end of the lex
			ret = _tcscspn(pbeg + 1, _T("]"));
			
			if (ret >= _tcslen(pbeg)) return FALSE; // sintax error
			memcpy(pLex, pbeg, ((ret + 2) * sizeof(TCHAR)));
			pLex[ret+2] = 0;
			EndInd += ret + 1;
			return TRUE;
		break;	
	case '"':
			pbeg = pstr + ret;
			EndL = '\"';
			EndInd = ret;

			// find end of the lex
			ret = _tcscspn( pbeg + 1, _T("\""));

			if (ret >= _tcslen(pbeg)) 
				return FALSE; // syntax error

			memcpy(pLex, pbeg, ((ret + 2) * sizeof(TCHAR)));

			pLex[ret+2] = 0;
			EndInd += ret + 1;

			return TRUE;
		break;
	case '=':
			pLex[0] = '='; pLex[1] = 0;
			EndInd = 1;
			return TRUE;
		break;
	case ':': // 
			if (!(pToken = _tcstok( pstr, _T(" :"))))
				return FALSE; // something wrong
			_tcscpy(pLex, pToken);
			EndInd += ret + 1;
			pToken[_tcslen(pToken)] = ':'; // restore 
			return TRUE;
	}
	return FALSE;
}

BOOL CADSparse::ParseLine(ParseLineStruct& pls, TCHAR *pLine)
{
	BOOL bRet;
 	int Ind, temp;
 	TCHAR *pLineCurr;

	bRet = GetLex(pLine, pls.m_KeyName, Ind);

	if (!bRet) 
		return FALSE;

	if (pls.m_KeyName[0] == '[') 
	{
		pls.m_EntryType = T_KEY;
	}
	else 
	{
		if (pls.m_KeyName[0] == '"') 
		{
			pls.m_EntryType = T_VALUE;
		}
		else 
		{
			return FALSE; // wrong syntax
		}
	}

	// delete the first and the last
	temp = _tcslen(pls.m_KeyName) - 2;
	memmove(pls.m_KeyName, pls.m_KeyName + 1, temp * sizeof(TCHAR));
	pls.m_KeyName[temp] = '\0';

	if (pls.m_EntryType == T_VALUE) // value name, move to m_ValName
	{
		memcpy(pls.m_ValName, pls.m_KeyName, (_tcslen(pls.m_KeyName) + 1) * sizeof(TCHAR));
	}
	else 
	{
		// key name
		return TRUE;
	}

	// we have just Value here
	pLineCurr = pLine + Ind + 1;

	// find the = sign
	bRet = GetLex(pLineCurr, pls.m_KeyName, Ind); // use pls.m_KeyName for temporary storage

	if (!bRet) 
	{
		return FALSE;
	}

	if (_memicmp(pls.m_KeyName, _T("="), (_tcslen(pls.m_KeyName) * sizeof(TCHAR))))
	{
		return FALSE; // syntax error
	}

	pLineCurr = pLineCurr + Ind;

	if(pLineCurr[0] != '"')
	{
		// not a string value
		bRet = GetLex(pLineCurr, pls.m_KeyName, Ind); 
		if (!bRet) 
		{
			return FALSE;
		}
	}
	else
	{
		_tcscpy(pls.m_KeyName, pLineCurr);
	}

	if (pls.m_KeyName[0] == '"')
	{
		pls.m_Type = REG_SZ;

		temp = _tcslen(pls.m_KeyName) * sizeof(TCHAR);
		_tcscpy((USHORT *)pls.m_Data, pls.m_KeyName + 1);
		pls.m_Data[temp - (2 * sizeof(TCHAR))] = 0;
		pls.m_DataSize = temp - (1 * sizeof(TCHAR)); // Null is included
		return TRUE;
	}

	if (_memicmp(pls.m_KeyName, _T("dword:"), (_tcslen(pls.m_KeyName) * sizeof(TCHAR))) == 0)
	{
		pls.m_Type = REG_DWORD;
		pLineCurr += Ind;
		// all value is hex!!!
        *(DWORD *)(&pls.m_Data)  = atox( pLineCurr );
		pls.m_DataSize = sizeof(DWORD);

		return TRUE;
	}

	/****************** HEX (REG_BINARY) DATA *******************/
	// hex data may span multiple lines.  Each dword of data is comma delimted
	// the '\' character is the line continuation character
	if (_memicmp(pls.m_KeyName, _T("hex:"), _tcslen(pls.m_KeyName) * sizeof(TCHAR)) == 0)
	{
		DWORD dd;
		BYTE *pDataCurr;
		TCHAR *pCurr;
		TCHAR *pToken ;

		pls.m_Type = REG_BINARY;
		pls.m_DataSize = 0;
		
		// move past the "hex:" identifier
		pLineCurr += Ind;

		pDataCurr = (BYTE *)pls.m_Data;
        pCurr = pLineCurr;

		if (pToken = _tcstok( pLineCurr, _T(", ")))
		{
			do 
			{
                pCurr = pToken; 

				if (_tcslen(pCurr) != 0)
				{
                    dd = atox(pCurr);

                    *pDataCurr++ = *(BYTE *)&dd;
					pls.m_DataSize++;
				}
			}while ((pToken = _tcstok( NULL, _T(", "))) != NULL);
		}

		return TRUE;
	}

	/****************** HEX(7) (REG_MULTI_SZ) DATA *******************/
	// hex data may span multiple lines.  Each dword of data is comma delimted
	// the '\' character is the line continuation character
	if (_memicmp(pls.m_KeyName, _T("hex(7):"), _tcslen(pls.m_KeyName) * sizeof(TCHAR)) == 0)
	{
		DWORD dd;
		BYTE *pDataCurr;
		TCHAR *pCurr;
		TCHAR *pToken ;

		pls.m_Type = REG_MULTI_SZ;
		pls.m_DataSize = 0;
		
		// move past the "hex:" identifier
		pLineCurr += Ind;

		pDataCurr = (BYTE *)pls.m_Data;
        pCurr = pLineCurr;

		if (pToken = _tcstok( pLineCurr, _T(", ")))
		{
			do 
			{
                pCurr = pToken; 

				if (_tcslen(pCurr) != 0)
				{
                    dd = atox(pCurr);

                    *pDataCurr++ = *(BYTE *)&dd;
					pls.m_DataSize++;

					// because CE wants the string as Unicode for insert...
                    dd = '\0';

                    *pDataCurr++ = *(BYTE *)&dd;
					pls.m_DataSize++;
				}
			}while ((pToken = _tcstok( NULL, _T(", "))) != NULL);
		}

		return TRUE;
	}

	/****************** multi_sz (REG_MULTI_SZ) DATA *******************/
	// this is an alterante form for REG_MULTI_SZ data to the standard hex(7)
	if (_memicmp(pls.m_KeyName, _T("multi_sz:"), (_tcslen(pls.m_KeyName) * sizeof(TCHAR))) == 0)
	{
		USHORT *p = (USHORT*)pls.m_Data;
		pls.m_Type = REG_MULTI_SZ;
		pls.m_DataSize = 0;

		// move past the "multi_sz:" identifier and the quote
		pLineCurr += (Ind + 1);

		// find any comma
		UINT commaAt = _tcscspn(pLineCurr, _T(","));

		do 
		{
			// copy the value, less the terminating quote
			_tcsncpy(p, pLineCurr, commaAt - 1);

			// null terminate
			p[(commaAt - 1)] = '\0';

			// add size (less terminating quote)
			pls.m_DataSize += ((commaAt - 1) * sizeof(TCHAR) + 1);

			// move past the comma
			pLineCurr += commaAt;

			// move to next quote (in case of spaces after comma)
			//if(commaAt < _tcslen(pLineCurr))
			if(pLineCurr[0] == NULL)
			{
				break;
			}
			while(pLineCurr[0] != '"')
			{
				pLineCurr++;
			}
			
			// move past quote
			pLineCurr++;

			// increment our data pointer
			p+= commaAt;
			commaAt = _tcscspn(pLineCurr, _T(","));
		}while(true);

		return TRUE;
	}

	/****************** HEX(2) (REG_EXPAND_SZ) DATA *******************/
	// hex data may span multiple lines.  Each dword of data is comma delimted
	// the '\' character is the line continuation character
	if (_memicmp(pls.m_KeyName, _T("hex(2):"), _tcslen(pls.m_KeyName) * sizeof(TCHAR)) == 0)
	{
		DWORD dd;
		BYTE *pDataCurr;
		TCHAR *pCurr;
		TCHAR *pToken ;

		pls.m_Type = REG_EXPAND_SZ;
		pls.m_DataSize = 0;
		
		// move past the "hex:" identifier
		pLineCurr += Ind;

		pDataCurr = (BYTE *)pls.m_Data;
        pCurr = pLineCurr;

		if (pToken = _tcstok( pLineCurr, _T(", ")))
		{
			do 
			{
                pCurr = pToken; 

				if (_tcslen(pCurr) != 0)
				{
                    dd = atox(pCurr);

                    *pDataCurr++ = *(BYTE *)&dd;
					pls.m_DataSize++;

					// because CE wants the string as Unicode for insert...
                    dd = '\0';

                    *pDataCurr++ = *(BYTE *)&dd;
					pls.m_DataSize++;
				}
			}while ((pToken = _tcstok( NULL, _T(", "))) != NULL);
		}

		return TRUE;
	}

	return TRUE;
}

CKeyEntry *CADSparse::ParseFile()
{
	TCHAR			Line[MAX_CAHRS_LINE];
 	int				ret;
 	ParseLineStruct pls;
 	CKeyEntry		*pKE		= NULL;
	CKeyEntry		*pKEHead	= NULL;
	CKeyEntry		*pKETail	= NULL;
 	CValueEntry		*pVE;
	int				token;
	TCHAR			lineBuffer[255];
	BOOL			continued = FALSE;

	if (!m_pFile) return FALSE;

	while (!(ret = m_pFile->GetNextLine(lineBuffer, sizeof(lineBuffer))))
	{
		continued = FALSE;
		Line[0] = '\0';

		TrimComment(lineBuffer);
		RTrim(lineBuffer);

		token = (_tcslen(lineBuffer) - 1);

		// look for line continuations ("\")
		while(_tcsncmp(&lineBuffer[token], _T("\\"), 1) == 0)
		{
			continued = TRUE;
			// continuation found, so build buffer until there isn't one
			lineBuffer[token] = '\0';
			_tcscat(Line, lineBuffer);
			m_pFile->GetNextLine(lineBuffer, sizeof(lineBuffer));
			token = (_tcslen(lineBuffer) - 1);
		}

		if(continued)
		{
			// grab last line
			_tcscat(Line, lineBuffer);
		}
		else
		{
			// copy the buffer to the Line
			_tcscpy(Line, lineBuffer);
		}
		
		if (!ParseLine(pls, Line)) 
            continue; // comment or it's not recognizable or empty

        if (pls.m_EntryType == T_KEY)
		{
			pKE = new CKeyEntry(pls.m_KeyName);
			if (!pKEHead) 
				pKETail = pKEHead = pKE;
			else
			{
				pKETail->m_pNext = pKE;
				pKETail = pKE;
			}
		}
		else
		{
			// Value entry
			pVE = new CValueEntry(pls.m_ValName, pls.m_Type, pls.m_Data, pls.m_DataSize);

			pKE->InsertValue(pVE);
			DEBUGMSG(TRUE, (_T("  Parsed: Value = %s Data = %s\r\n"), pls.m_ValName, pls.m_Data));
		}
	}

	if (ret == 1)
		return pKEHead;
	else
		return NULL;
	
}

// this function try to update Registry even it is failed for some keys
//	returns FALSE if it was failed on some keys
BOOL CADSparse::UpdateRegistry(CKeyEntry *pKeyList)
{
	CValueEntry	*pVList;
 	TCHAR	*pch;
 	LONG	lRet;	
 	HKEY	hRootKey, hTargetKey;
 	DWORD	dwDisposition;
 	DWORD	nErrorCount = 0;
 	DWORD  uSize;



	while (pKeyList)
	{
		// extract the root key
		pch = _tcschr( pKeyList->m_pKeyName, '\\');
		if (!pch) {
            pKeyList = pKeyList->m_pNext;
            nErrorCount++; 
            continue;
        } // file is corrupted
		*pch = 0;

		if (!_tcscmp(pKeyList->m_pKeyName,_T("HKEY_LOCAL_MACHINE"))) hRootKey = HKEY_LOCAL_MACHINE;
		if (!_tcscmp(pKeyList->m_pKeyName,_T("HKEY_CLASSES_ROOT"))) hRootKey = HKEY_CLASSES_ROOT;
		if (!_tcscmp(pKeyList->m_pKeyName,_T("HKEY_CURRENT_USER"))) hRootKey = HKEY_CURRENT_USER;


		lRet = RegCreateKeyEx(	hRootKey,
								pch + 1,			// subkey
								0,				// reserved
								NULL,			// class name
                                REG_OPTION_NON_VOLATILE,
								KEY_ALL_ACCESS ,// actually
								NULL,			// security attributes
								&hTargetKey,	// key to open
								&dwDisposition	// disposition
							);

		DEBUGMSG (true, (_T("Creating key: %s\r\n"), pch + 1));
        DEBUGMSG (true, (L"CADSparse:UpdateRegistry:RegCreateKeyEx lRet = %d\r\n",lRet));

		if (lRet != ERROR_SUCCESS)
		{
            RETAILMSG (true, (L"CADSparse::UpdateRegistry::RegCreateKeyEx err = %d\r\n",
                                GetLastError()));

            pVList = pVList->m_pNext;
            nErrorCount++; continue;
        } // something wrong, try other keys

		pVList = pKeyList->m_pValueList;
		while (pVList)
		{
			if (pVList->m_Type == REG_SZ)
            {
				DEBUGMSG (true, (_T("Writing Value (%s): %s\r\n"), pVList->m_pValName,  pVList->m_pData));
				uSize = (_tcslen((TCHAR*)pVList->m_pData) + 1) * sizeof(TCHAR);
            }
            else
            {
				DEBUGMSG (true, (_T("Writing Value (%s): %i\r\n"), pVList->m_pValName,  pVList->m_pData));
                uSize = pVList->m_DataSize;
            }

			lRet = RegSetValueEx(	hTargetKey,			// key handle
									pVList->m_pValName,               // Value name
									0,					// reserved
									pVList->m_Type,		// type
									pVList->m_pData,	// data
									uSize           	// data size
								);

            pVList = pVList->m_pNext;

            if (lRet != ERROR_SUCCESS) nErrorCount++;  // something wrong, try other values
		}

		RegCloseKey( hTargetKey ); 

        pKeyList = pKeyList->m_pNext;
	}

	return TRUE;
}



DWORD   CADSparse::atox(TCHAR *pstr)
{
	DWORD result = 0;
    // skip initial garbage 

    while(*pstr <= ' ' || *pstr == '0' || *pstr == 'x') 
	{
        if(!*pstr) return result; 
        pstr++;
    }

    while(*pstr) 
	{
        if(is_digit(*pstr)) 
        {
            result = (result << 4) + (*pstr - '0'); 
            pstr++;
            continue;
        }

        if(is_xdigit(*pstr)) 
        { 
            result = (result << 4) + (tolower(*pstr) - 'a' + 10); 
            pstr++; 
            continue;
        }

        break;
    }
    return result;
}

void TrimComment(TCHAR *szVal)
{
	// clean any trailing white space
	RTrim(szVal);

	// reverse the string
	_tcsrev(szVal);

	// find the next quote
	UINT quoteAt = _tcscspn(szVal, _T("\""));
	
	// find the semicolon
	UINT semiAt = _tcscspn(szVal, _T(";"));

	_tcsrev(szVal);

	// strip any comment
	if((semiAt < _tcslen(szVal)) && (quoteAt > semiAt))
	{
		szVal[_tcslen(szVal) - semiAt - 1] = '\0';
		RTrim(szVal);
	}
}

void RTrim(TCHAR *szVal)
{
	// trim trailing spaces and tabs
	for(int i = _tcslen(szVal) - 1 ; i >= 0 ;)
	{
		if((szVal[i] == 0x20) || (szVal[i] == 0x09))
		{
			szVal[i] = '\0';
			i--;
		}
		else
		{
			break;
		}
	}
}
